#!/usr/bin/env python

import logging
import csv
import os
import re
from abc import ABCMeta, abstractmethod

from soyuz.dx.variables import Type


class DataFile(object):
    __metaclass__ = ABCMeta

    def __init__(self, location, name, seq_folder):
        self.__location = location
        self.__name = name
        self.__seq_folder = seq_folder

    def get_location(self):
        return self.__location

    def get_name(self):
        return self.__name

    @property
    def is_valid(self):
        return bool(self.get_regex().match(self.get_name()))

    def get_seq_folder_name(self):
        return self.__seq_folder.get_name()

    def get_relative_path(self):
        relpath = os.path.relpath(self.get_location(), self.__seq_folder.get_path())
        if relpath == ".":
            return ""
        return relpath

    def get_full_path(self):
        return os.path.join(self.__location, self.__name)

    @abstractmethod
    def get_type(self):
        raise NotImplementedError()

    @abstractmethod
    def get_regex(self):
        raise NotImplementedError()

    @abstractmethod
    def get_sample_id(self):
        raise NotImplementedError()


class RawDataFile(DataFile):
    def get_sample_id(self):
        return None

    def get_type(self):
        return None

    def get_regex(self):
        return re.compile(".+?")


class WesDataFile(DataFile):
    def get_regex(self):
        return re.compile("(?P<sample_id>[A-Za-z0-9-.]*)_[A-Za-z0-9-_]*.(?P<extension>fastq.gz|fastq|bam)")

    def get_type(self):
        m = self.get_regex().match(self.get_name())
        if m:
            extension = m.group("extension")
            if extension == "fastq.gz" or extension == "fastq":
                return Type.FASTQ
            if extension == "bam":
                return Type.BAM
        return None

    def get_sample_id(self):
        m = self.get_regex().match(self.get_name())
        if m:
            return m.group("sample_id")
        return None


class PersonalisDataFile(DataFile):
    def get_regex(self):
        return re.compile("[A-Za-z0-9-]*_(?P<sample_id>[A-Za-z0-9-.]*)_[A-Za-z0-9-_.]*.(?P<extension>fastq.gz|fastq|bam)")

    def get_type(self):
        m = self.get_regex().match(self.get_name())
        if m:
            extension = m.group("extension")
            if extension == "fastq.gz" or extension == "fastq":
                return Type.FASTQ
            if extension == "bam":
                return Type.BAM
        return None

    def get_sample_id(self):
        m = self.get_regex().match(self.get_name())
        if m:
            return m.group("sample_id")
        return None


class QcDataFile(DataFile):
    @property
    def is_valid(self):
        if not super(QcDataFile, self).is_valid:
            return False

        if self.get_type() == Type.CSV:
            sniffer = csv.Sniffer()
            with open(self.get_full_path(), 'r') as f:
                data = f.read()
            if sniffer.sniff(data).delimiter != ',':
                logging.error("CSV file {}: expected a comma delimiter, but actual is {}.".format(self.get_full_path(), sniffer.sniff(data).delimiter))
                return False
            if not sniffer.has_header(data):
                logging.error("CSV file {} doesn't have a header.".format(self.get_full_path()))
                return False
            with open(self.get_full_path(), 'r') as f:
                for line in csv.DictReader(f):
                    if not re.compile("^[a-zA-Z0-9\\-_.]+$").match(line['Natera Sample ID']):
                        logging.error("CSV file {} doesn't have a valid 'Natera Sample ID'.".format(self.get_full_path()))
                        return False
                    if not line['aveDedupOnTarget']:
                        logging.error("CSV file {} doesn't have an 'aveDedupOnTarget'.".format(self.get_full_path()))
                        return False
                    metrics = ["Total recovery (ug)", "%gDNA", "A260/A280", "Library into Capture (ng)", "PreCap BioA size (bp)", "Post capture Library concentration (ng/uL)", "PostCap BioA size (bp)", "totReads", "totPairedReads", "averageBaseQ", "combinedQ30pct", "r1Q30pct", "r2Q30pct", "mapq", "pctAlignedToGenome", "gcPCT", "pctDuplicate", "pctOnPanel"]
                    for key, val in line.items():
                        if key in metrics and not val:
                            logging.warning("CSV file {} doesn't have a metric {} defined.".format(self.get_full_path(), key))
                    if line["% Tumor Fraction"] and not line["ffpe100x"]:
                        logging.warning("CSV file {} doesn't have 'ffpe100x' defined.".format(self.get_full_path()))
                    if not line["% Tumor Fraction"] and not line["norm30x"]:
                        logging.warning("CSV file {} doesn't have 'norm30x' defined.".format(self.get_full_path()))
        return True

    def get_regex(self):
        return re.compile("[A-Za-z0-9-_ ]*.(?P<extension>csv|pdf|xlsx)")

    def get_type(self):
        m = self.get_regex().match(self.get_name())
        if m:
            extension = m.group("extension")
            if extension == "csv":
                return Type.CSV
            if extension == "pdf":
                return Type.PDF
            if extension == "xlsx":
                return Type.XLSX
        return None

    def get_sample_id(self):
        return None


class PersonalisQcDataFile(QcDataFile):
    @property
    def is_valid(self):
        if not super(QcDataFile, self).is_valid:
            return False

        if self.get_type() == Type.CSV:
            sniffer = csv.Sniffer()
            with open(self.get_full_path(), 'r') as f:
                data = f.read()
            if sniffer.sniff(data).delimiter != ',':
                logging.error("CSV file {}: expected a comma delimiter, but actual is {}.".format(self.get_full_path(), sniffer.sniff(data).delimiter))
                return False
            if not sniffer.has_header(data):
                logging.error("CSV file {} doesn't have a header.".format(self.get_full_path()))
                return False
            with open(self.get_full_path(), 'r') as f:
                for line in csv.DictReader(f):
                    if not re.compile("^[a-zA-Z0-9\\-_.]+$").match(line['Natera Sample ID']):
                        logging.error("CSV file {} doesn't have a valid 'Natera Sample ID'.".format(self.get_full_path()))
                        return False
                    if not line['aveDedupOnTarget']:
                        logging.error("CSV file {} doesn't have an 'aveDedupOnTarget'.".format(self.get_full_path()))
                        return False
                    metrics = ["% Tumor Fraction", "extraction_ng", "DIN", "Q30_reads", "mapQ", "pct_reads_mapped", "pct_GC", "duplication_rate", "cap_spec", "aveDedupOnTarget", "f_above_30x", "f_above_100x"]
                    for key, val in line.items():
                        if key in metrics and not val:
                            logging.warning("CSV file {} doesn't have a metric {} defined.".format(self.get_full_path(), key))
        return True



#!/usr/bin/env python

import logging
import os
import re
import time
from abc import ABCMeta, abstractmethod

from soyuz.data.files import WesDataFile, RawDataFile, QcDataFile, PersonalisDataFile, PersonalisQcDataFile
from soyuz.utils import UploaderException


class WatchDirectory(object):
    UPLOADED_FILE = ".uploaded"

    def __init__(self, path):
        self.__path = path

    def is_uploaded(self, folder):
        return str(folder) in self.__read_file(self.__uploaded_file)

    def complete(self, folder):
        self.__write_file(self.__uploaded_file, str(folder))

    @property
    def __uploaded_file(self):
        return os.path.join(self.__path, self.UPLOADED_FILE)

    @property
    def __copied_file(self):
        return os.path.join(self.__path, self.COPIED_FILE)

    def get_seq_folders(self):
        return self.__get_seq_folders(self.__path)

    def __get_seq_folders(self, root):
        result = []
        if not os.path.exists(root):
            logging.error("Watch directory {} doesn't exist".format(root))
            return result

        for f in os.listdir(root):
            path = os.path.join(root, f)
            if os.path.isdir(path):
                seq_folder = SeqFolderFactory.create(root, f)
                if seq_folder:
                    result.append(seq_folder)
                else:
                    result += self.__get_seq_folders(path)
        return result

    @staticmethod
    def __read_file(file_with_history):
        result = []
        if not os.path.isfile(file_with_history):
            return result
        with open(file_with_history, "r") as f:
            for line in f.readlines():
                result.append(line.strip())
        return result

    @staticmethod
    def __write_file(file_with_history, seq_folder_name):
        with open(file_with_history, "a") as f:
            f.write(seq_folder_name)


class SeqFolderBase(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, path):
        self._path = self.__get_abs_path(path).replace("\\", "/")

    def get_name(self):
        return os.path.basename(self._path)

    def get_path(self):
        return self._path

    @property
    @abstractmethod
    def is_valid(self):
        raise NotImplementedError()

    @property
    def is_uploading(self):
        return os.path.isfile(self._path + ".uploading")

    @property
    @abstractmethod
    def is_completed(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def complete(self):
        raise NotImplementedError()

    def is_older_than(self, seconds):
        return time.time() - os.path.getmtime(self.__str__()) > seconds

    @abstractmethod
    def list_files(self):
        raise NotImplementedError()

    @staticmethod
    def __get_abs_path(path):
        abs_path = os.path.abspath(os.path.expanduser(path))
        if not os.path.isdir(abs_path):
            raise UploaderException("{} does not exist".format(abs_path))
        return abs_path


class WesSignateraSeqFolder(SeqFolderBase):
    FOLDER_REGEX = re.compile("[A-Z0-9]*_[0-9]{8}")

    def __init__(self, path):
        SeqFolderBase.__init__(self, path)
        self.__wes_subfolder = os.path.join(self._path, "WES_data")
        self.__qc_subfolder = os.path.join(self._path, "QC_reports")

    @property
    def is_valid(self):
        if WesSignateraSeqFolder.FOLDER_REGEX.match(self.get_name()) is None \
                or not os.path.isdir(self.__wes_subfolder) \
                or not os.path.isdir(self.__qc_subfolder):
            return False

        if len(self.__list_from_wes()) == 0:
            logging.info("{} does not contain any data".format(self.__wes_subfolder))
            return False
        if len(self.__list_from_qc_reports()) == 0:
            logging.info("{} does not contain any data".format(self.__qc_subfolder))
            return False

        success = True
        for f in self.list_files():
            if not f.is_valid:
                logging.info("{} has invalid format".format(f.get_full_path()))
                success = False
        return success

    def list_files(self):
        return self.__list_from_wes() + self.__list_from_qc_reports()

    def __list_from_wes(self):
        result = []
        for f in os.listdir(self.__wes_subfolder):
            result.append(WesDataFile(self.__wes_subfolder, f, self))
        return result

    def __list_from_qc_reports(self):
        result = []
        for f in os.listdir(self.__qc_subfolder):
            result.append(QcDataFile(self.__qc_subfolder, f, self))
        return result


class PersonalisSeqFolder(SeqFolderBase):
    FOLDER_REGEX = re.compile("[A-Za-z0-9_-]{1,48}")

    def __init__(self, path):
        SeqFolderBase.__init__(self, path)
        self.__alignment_subfolder = os.path.join(self._path, '{}', "Alignment")
        self.__fastq_subfolder = os.path.join(self._path, '{}', "FASTQ")
        self.__qc_metrics_path = self._path

    @property
    def is_valid(self):
        if PersonalisSeqFolder.FOLDER_REGEX.match(self.get_name()) is None:
            return False

        if len(self.__list_qc_metrics(self._path)) == 0:
            logging.info("{} does not contain any data".format(self.__qc_metrics_path))
            return False

        for folder in os.listdir(self._path):
            if not os.path.isdir(os.path.join(self._path, folder)):
                continue
            if not os.path.isdir(self.__alignment_subfolder.format(folder)):
                return False
            if len(self.__list_from(self.__alignment_subfolder.format(folder))) == 0:
                logging.info("{} does not contain any data".format(self.__alignment_subfolder))
                return False
            if os.path.isdir(self.__fastq_subfolder.format(folder)) and \
                    len(self.__list_from(self.__fastq_subfolder.format(folder))) == 0:
                logging.info("{} does not contain any data".format(self.__fastq_subfolder))
                return False

        success = True
        for f in self.list_files():
            if not f.is_valid:
                logging.info("{} has invalid format".format(f.get_full_path()))
                success = False
        return success

    def list_files(self):
        list_files = self.__list_qc_metrics(self._path)
        for folder in os.listdir(self._path):
            if not os.path.isdir(os.path.join(self._path, folder)):
                continue
            list_files.extend(self.__list_from(self.__alignment_subfolder.format(folder)))
            if os.path.isdir(self.__fastq_subfolder.format(folder)):
                list_files.extend(self.__list_from(self.__fastq_subfolder.format(folder)))
        return list_files

    def __list_from(self, subfolder):
        result = []
        for f in os.listdir(subfolder):
            result.append(PersonalisDataFile(subfolder, f, self))
        return result

    def __list_qc_metrics(self, path):
        result = []
        for f in os.listdir(path):
            if os.path.isfile(os.path.join(path, f)):
                result.append(PersonalisQcDataFile(path, f, self))
        return result


class RawSeqFolder(SeqFolderBase):
    def __init__(self, path):
        super(RawSeqFolder, self).__init__(path)

    @property
    def is_valid(self):
        return True

    def list_files(self):
        result = []
        for root, subdirs, files in os.walk(self._path):
            for f in files:
                result.append(RawDataFile(root, f, self))
        return result


class BgiSeqFolder(SeqFolderBase):
    FOLDER_REGEX = re.compile("V[A-Za-z0-9]{9}")

    def __init__(self, path):
        super(BgiSeqFolder, self).__init__(path)
        self.__lane_folders = [os.path.join(self._path, lane) for lane in ["L01", "L02", "L03", "L04"]]
        self.__wes_subfolder = os.path.join(self._path, "WES_data")
        self.__qc_subfolder = os.path.join(self._path, "QC_reports")

    @property
    def is_completed(self):
        if self.is_valid:
            return all([os.path.isfile(os.path.join(lane_folder, "{}_{}_FileInfo.csv".format(os.path.basename(self._path), os.path.basename(lane_folder)))) for lane_folder in self.__lane_folders])
        return False

    @property
    def is_valid(self):
        if (BgiSeqFolder.FOLDER_REGEX.match(self.get_name())
                and all([os.path.join(self._path, folder) in self.__lane_folders for folder in os.listdir(self._path) if os.path.isdir(os.path.join(self._path, folder))])):
            return True
        return False

    def list_files(self):
        result = []
        for root, subdirs, files in os.walk(self._path):
            for f in files:
                result.append(RawDataFile(root, f, self))
        return result


class SeqFolderFactory(object):
    @staticmethod
    def create(base_path, name):
        if BgiSeqFolder.FOLDER_REGEX.match(name):
            return BgiSeqFolder(os.path.join(base_path, name))
        if WesSignateraSeqFolder.FOLDER_REGEX.match(name):
            return WesSignateraSeqFolder(os.path.join(base_path, name))
        logging.error("Folder {} is not allowed.".format(os.path.join(base_path, name)))
        return None

    @staticmethod
    def create_from_full_path(full_path):
        base_path, name = SeqFolderFactory.__parse_full_path(full_path)
        return SeqFolderFactory.create(base_path, name)

    @staticmethod
    def __parse_full_path(full_path):
        path = full_path.rstrip('/')
        return os.path.dirname(path), os.path.basename(path)



#!/usr/bin/env python

import argparse
import json
import logging
import multiprocessing
import os
import sys

from soyuz import __version__ as version
from soyuz.utils import UploaderException, Logging
from soyuz.data.storages import StorageType


class Parameters(object):
    UPLOAD_ACTION = "upload"
    WATCH_ACTION = "watch"
    WATCH_INTERVAL = 600
    WES_SIGNATERA_UPLOAD = "wes_signatera"
    RAW_UPLOAD = "raw"
    BGI_UPLOAD = "bgi"
    PERSONALIS_UPLOAD = 'personalis'
    CONFIG_ACTION = "config"
    GET_CONFIG = "get"
    SET_CONFIG = "set"
    START_WATCH = "start"
    STOP_WATCH = "stop"
    STATUS_WATCH = "status"
    VERSION_ACTION = "version"

    def __init__(self, args=sys.argv[1:]):
        parser = argparse.ArgumentParser(description="Constellation Sequencing Uploader v{}".format(version))

        parser.add_argument('--token', '-t', help='DNAnexus token with access to a single project with level UPLOAD')

        sp = parser.add_subparsers(dest='command')
        sp.required = True

        ##################
        # uploader upload
        ##################
        upload_parser = sp.add_parser(Parameters.UPLOAD_ACTION, help='Upload sequencing folder')
        upload_parser.set_defaults(action=Parameters.UPLOAD_ACTION)

        sup = upload_parser.add_subparsers(dest='upload type')
        sup.required = True

        # uploader upload wes_signatera
        wes_signatera_parser = sup.add_parser(Parameters.WES_SIGNATERA_UPLOAD, help='Upload for WES Signatera')
        wes_signatera_parser.set_defaults(upload_type=Parameters.WES_SIGNATERA_UPLOAD)
        wes_signatera_parser.add_argument('folder', metavar='Folder')

        # uploader upload bgi
        bgi_parser = sup.add_parser(Parameters.BGI_UPLOAD, help='Upload for BGI with fastq files')
        bgi_parser.set_defaults(upload_type=Parameters.BGI_UPLOAD)
        bgi_parser.add_argument('folder', metavar='Folder')

        # uploader upload raw
        raw_parser = sup.add_parser(Parameters.RAW_UPLOAD, help='Raw upload without any validation and modifications')
        raw_parser.set_defaults(upload_type=Parameters.RAW_UPLOAD)
        raw_parser.add_argument('folder', metavar='Folder')

        # uploader upload personalis
        personalis_parser = sup.add_parser(Parameters.PERSONALIS_UPLOAD, help='Upload for Personalis')
        personalis_parser.set_defaults(upload_type=Parameters.PERSONALIS_UPLOAD)
        personalis_parser.add_argument('folder', metavar='Folder')

        ##################
        # uploader watch
        ##################
        watch_parser = sp.add_parser(Parameters.WATCH_ACTION, help='Watch sequencing folder')
        watch_parser.set_defaults(action=Parameters.WATCH_ACTION)

        watch_parser.add_argument('--interval', '-n', default=Parameters.WATCH_INTERVAL, type=int, help='Interval in seconds between every run in watch mode')

        sup = watch_parser.add_subparsers(dest='watch type')
        sup.required = True

        # uploader watch start
        start_watch_parser = sup.add_parser(Parameters.START_WATCH, help='Start watching folder')
        start_watch_parser.set_defaults(watch_type=Parameters.START_WATCH)
        start_watch_parser.add_argument('folder', metavar='Folder')
        start_watch_parser.add_argument('--foreground', '-fg', help='Run in foreground.', default=False, action='store_true')

        # uploader watch stop
        stop_watch_parser = sup.add_parser(Parameters.STOP_WATCH, help='Stop watching folder')
        stop_watch_parser.set_defaults(watch_type=Parameters.STOP_WATCH)

        # uploader watch status
        status_watch_parser = sup.add_parser(Parameters.STATUS_WATCH, help='Status of watching folder')
        status_watch_parser.set_defaults(watch_type=Parameters.STATUS_WATCH)

        ##################
        # uploader config
        ##################
        config_parser = sp.add_parser(Parameters.CONFIG_ACTION,
                                      help='Configure uploader. This action updates ~/.uploader file')
        config_parser.set_defaults(action=Parameters.CONFIG_ACTION, show_help=config_parser.print_help)

        scp = config_parser.add_subparsers(dest='config action')
        scp.required = True

        # uploader config get
        get_config_parser = scp.add_parser(Parameters.GET_CONFIG,
                                           help='Get uploader configuration')
        get_config_parser.set_defaults(config_action=Parameters.GET_CONFIG)
        get_config_parser.add_argument('config_parameter_key', metavar='Key')

        # uploader config set
        set_config_parser = scp.add_parser(Parameters.SET_CONFIG,
                                           help='Configure uploader. This action updates ~/.uploader file')
        set_config_parser.set_defaults(config_action=Parameters.SET_CONFIG)
        set_config_parser.add_argument('config_parameter_key', metavar='Key')
        set_config_parser.add_argument('config_parameter_value', metavar='Value')

        ##################
        # uploader version
        ##################
        version_parser = sp.add_parser(Parameters.VERSION_ACTION, help='Show version')
        version_parser.set_defaults(action=Parameters.VERSION_ACTION)

        self.__args = parser.parse_args(args)

        if self.__args.action == Parameters.VERSION_ACTION:
            print("v" + version)
            quit(0)

    def get_token(self):
        return self.__args.token

    def get_interval(self):
        return self.__args.interval

    def get_folder(self):
        return self.__args.folder

    def foreground(self):
        return self.__args.foreground

    def get_action(self):
        return self.__args.action

    def get_upload_type(self):
        return self.__args.upload_type

    def get_watch_type(self):
        return self.__args.watch_type

    def get_config_action(self):
        return self.__args.config_action

    def get_config_parameter_key(self):
        return self.__args.config_parameter_key

    def get_config_parameter_value(self):
        return self.__args.config_parameter_value


class Settings(object):
    SETTING_FILE = ".uploader"

    def __init__(self, settings_base_path="~"):
        self.__settings = {}
        self.logging = Logging()
        self.__full_path_to_settings = os.path.join(os.path.expanduser(settings_base_path), Settings.SETTING_FILE)

        if os.path.isfile(self.__full_path_to_settings):
            try:
                self.__settings = json.load(open(self.__full_path_to_settings))
            except Exception:
                raise UploaderException("{} is not a valid JSON".format(Settings.SETTING_FILE))

    def get_token(self):
        return self.__get_dnanexus().get('token')

    def set_token(self, token):
        self.__get_dnanexus()["token"] = token

    def get_base_dir(self):
        return self.__get_dnanexus().get('basedir', '/data/seq')

    def set_basedir(self, basedir):
        self.__get_dnanexus()["basedir"] = basedir

    def get_storage(self):
        return self.__settings.get("storage", StorageType.DEFAULT)

    def set_storage(self, storage):
        if storage not in StorageType.ALL:
            raise UploaderException("Storage type {} is not available.".format(storage))
        self.__settings["storage"] = storage

    def get_exodus_url(self):
        return self.__get_exodus().get("url")

    def set_exodus_url(self, exodus_url):
        self.__get_exodus()["url"] = exodus_url

    def get_exodus_username(self):
        return self.__get_exodus().get("username")

    def set_exodus_username(self, exodus_username):
        self.__get_exodus()["username"] = exodus_username

    def get_exodus_password(self):
        return self.__get_exodus().get("password")

    def set_exodus_password(self, exodus_password):
        self.__get_exodus()["password"] = exodus_password

    def get_interval(self):
        return self.__settings.get("interval")

    def set_interval(self, interval):
        self.__settings["interval"] = int(interval)

    def get_ua_path(self):
        return self.__get_dnanexus().get("ua_path") if os.path.isfile(self.__get_dnanexus().get("ua_path", '')) else None

    def set_ua_path(self, ua_path):
        if not os.path.isfile(ua_path):
            raise UploaderException("ua_path {} no such file.".format(ua_path))
        self.__get_dnanexus()["ua_path"] = ua_path

    def get_ua_parameters(self):
        return self.__get_dnanexus().get("ua_parameters", '')

    def set_ua_parameters(self, ua_parameters):
        self.__get_dnanexus()["ua_parameters"] = ua_parameters

    def get_chunk_size(self):
        return int(self.__settings.get("chunk_size", 200*10**6))

    def set_chunk_size(self, chunk_size):
        if isinstance(chunk_size, basestring):
            if chunk_size.endswith('M'):
                chunk_size = int(chunk_size.rstrip('M')) * 10**6
            elif chunk_size.endswith('G'):
                chunk_size = int(chunk_size.rstrip('G')) * 10**9
            if chunk_size < 5*10**6 or chunk_size > 5*10**9:
                raise UploaderException("Chunk size should be between 5M and 5G")
        self.__settings["chunk_size"] = int(chunk_size)

    def get_process_count(self):
        return self.__settings.get("process_count", 1)

    def set_process_count(self, process_count):
        process_count = int(process_count)
        if process_count > multiprocessing.cpu_count():
            logging.info("The number of upload threads cannot be higher than the number of CPU cores. Setting to {}".format(multiprocessing.cpu_count()))
            process_count = multiprocessing.cpu_count()
        elif process_count == 0:
            logging.info("The number of upload threads cannot be lower than 1. Setting to 1")
            process_count = 1
        if self.get_storage() == StorageType.DNANEXUS and not self.get_ua_path():
            logging.info("The multiprocessing is not allowed for DNAnexus without defining ua_path. Setting to 1")
            process_count = 1
        self.__settings["process_count"] = process_count

    def dump(self):
        with open(self.__full_path_to_settings, 'wt') as out:
            json.dump(self.__settings, out, sort_keys=True, indent=4, separators=(',', ': '))

    def __get_dnanexus(self):
        return self.__settings.setdefault("dnanexus", {})

    def __get_exodus(self):
        return self.__settings.setdefault("exodus", {})



#!/usr/bin/env python

from abc import ABCMeta, abstractmethod
import logging
import os
import requests
import subprocess

import dxpy

from soyuz.utils import UploaderException, read_in_chunks


class Storage(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        self.set_context()

    @abstractmethod
    def set_context(self):
        raise NotImplementedError()

    @abstractmethod
    def validate_target_dir(self):
        raise NotImplementedError()

    @abstractmethod
    def upload_file(self):
        raise NotImplementedError()


class DNAnexusStorage(Storage):
    NAME = "dnanexus"

    def __init__(self, settings):
        self.basedir = settings.get_base_dir()
        self.token = settings.get_token()
        self.ua_path = settings.get_ua_path()
        self.ua_parameters = settings.get_ua_parameters()
        super(DNAnexusStorage, self).__init__()

    def set_context(self):
        if not self.token:
            raise UploaderException("Token was not specified")
        dxpy.set_security_context({'auth_token_type': 'Bearer', 'auth_token': self.token})
        projects = self.__get_projects()
        size = len(projects)
        if size == 0 or size > 1:
            raise UploaderException("Auth Token must have access to exactly 1 project with UPLOAD permission.")
        self.__project = projects[0]
        dxpy.set_project_context(self.__project)
        dxpy.set_workspace_id(self.__project)

    def get_project_id(self):
        return self.__project

    def get_project(self):
        return dxpy.DXProject(self.get_project_id())

    def get_file_by_id(self, file_id):
        return dxpy.get_handler(file_id)

    @staticmethod
    def __get_projects():
        result = []
        try:
            for project in dxpy.bindings.search.find_projects(level='UPLOAD'):
                result.append(str(project['id']))
        except dxpy.exceptions.InvalidAuthentication:
            pass
        return result

    def validate_target_dir(self, folder):
        if not folder.is_valid:
            raise UploaderException(
                    "{} is not valid".format(folder.get_name()))
        project = dxpy.DXProject(self.get_project_id())
        try:
            entities = project.list_folder(os.path.join(self.basedir, folder.get_name()))
            if len(entities["objects"]) > 0 or len(entities["folders"]) > 0:
                raise UploaderException(
                    "{} already exists under {}".format(folder.get_name(), self.basedir))
        except dxpy.exceptions.ResourceNotFound:
            pass

    def upload_file(self, data_file, remote_folder, types, properties):
        logging.info("Uploading {} to {}".format(data_file.get_full_path(), remote_folder))
        dx_file = dxpy.upload_local_file(data_file.get_full_path(),
                                         folder=remote_folder,
                                         keep_open=True,
                                         parents=True)
        if dx_file:
            dx_file.add_types(types)
            dx_file.set_properties(properties)
            dx_file.close()
        else:
            raise UploaderException("Failed to upload {}".format(data_file.get_full_path()))
        return dx_file.get_id()


class DNAnexusStorageUA(DNAnexusStorage):
    def upload_file(self, data_file, remote_folder, types, properties):
        logging.info("Uploading {} to {}".format(data_file.get_full_path(), remote_folder))
        args = [r'"{}"'.format(self.ua_path), data_file.get_full_path().replace("(", "\(").replace(")", "\)")]
        args.extend(["--auth-token", self.token])
        args.extend(["-p", self.get_project_id()])
        args.extend(["-f", remote_folder])
        args.extend([self.ua_parameters])
        args.extend(["--type {}".format(_type) for _type in types])
        args.extend(["--property {}={}".format(key, val) for key, val in properties.items()])
        file_id = subprocess.check_output(" ".join(args), shell=True).strip()
        return file_id


class ExodusStorage(Storage):
    NAME = "exodus"

    def __init__(self, settings):
        self.basedir = settings.get_base_dir()
        self.exodus_url = settings.get_exodus_url()
        self.exodus_username = settings.get_exodus_username()
        self.exodus_password = settings.get_exodus_password()
        self.chunk_size = settings.get_chunk_size()
        super(ExodusStorage, self).__init__()

    def set_context(self):
        if not self.exodus_url:
            raise UploaderException("Exodus URL was not specified")

    def validate_target_dir(self, folder):
        if not folder.is_valid:
            raise UploaderException(
                    "{} is not valid".format(folder.get_name()))

    def upload_file(self, data_file, remote_folder, types, properties):
        logging.info("Uploading {} to {}".format(data_file.get_full_path(), remote_folder))
        data = {'name': data_file.get_name(),
                'folder': remote_folder,
                'types': types,
                'properties': properties,
                'tags': []}
        response = requests.post(os.path.join(self.exodus_url, "2.0/data/files/new"), json=data, auth=(self.exodus_username, self.exodus_password))
        response.raise_for_status()
        data = response.json()
        file_id = data['file']['link']
        uploadId = data['uploadId']

        with open(data_file.get_full_path(), 'rb') as file_object:
            parts = []
            for i, chunk in enumerate(read_in_chunks(file_object, self.chunk_size)):
                data = {'file': {
                            'link': file_id},
                        'uploadId': uploadId,
                        'part': i+1}
                response = requests.post(os.path.join(self.exodus_url, "2.0/data/files/part-upload"), json=data, auth=(self.exodus_username, self.exodus_password))
                response.raise_for_status()
                data = response.json()
                uploadUrl = data['uploadUrl']

                response = requests.put(uploadUrl, data=chunk)
                response.raise_for_status()
                headers = response.headers
                etag = headers['ETag']
                parts.append({'eTag': etag,
                              'part': i+1})

        data = {'file': {
                    'link': file_id},
                'uploadId': uploadId,
                'contentLength': os.path.getsize(data_file.get_full_path()),
                'parts': parts
                }
        response = requests.post(os.path.join(self.exodus_url, "2.0/data/files/finish-upload"), json=data, auth=(self.exodus_username, self.exodus_password))
        response.raise_for_status()
        data = response.json()
        file_id = data['link']
        logging.info("Successfully uploaded: {} to {}".format(data_file.get_full_path(), remote_folder))
        return file_id


class StorageType(object):
    DNANEXUS = DNAnexusStorage.NAME
    EXODUS = ExodusStorage.NAME
    DEFAULT = DNANEXUS
    ALL = [DNANEXUS, EXODUS]


class StorageFactory(object):
    @staticmethod
    def create(settings):
        if settings.get_storage() == DNAnexusStorage.NAME:
            return DNAnexusStorageUA(settings) if settings.get_ua_path() else DNAnexusStorage(settings)
        elif settings.get_storage() == ExodusStorage.NAME:
            return ExodusStorage(settings)
        raise UploaderException("Storage with name {} was not found".format(settings.get_storage()))



#!/usr/bin/env python

import logging
import multiprocessing
import os
import time
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager

from soyuz.dx.sentinels import WesSignateraSentinel, RawSentinel, PersonalisSentinel
from soyuz.dx.variables import Type, Property
from soyuz.data.folders import WesSignateraSeqFolder, PersonalisSeqFolder, RawSeqFolder, BgiSeqFolder
from soyuz.data.storages import StorageFactory
from soyuz.utils import UploaderException


@contextmanager
def poolcontext(*args, **kwargs):
    pool = multiprocessing.Pool(*args, **kwargs)
    yield pool
    pool.terminate()


class WatchUploader(object):
    __metaclass__ = ABCMeta

    def __init__(self, settings):
        self.settings = settings

    def watch(self, watch_dir):
        while True:
            for seq_folder in watch_dir.get_seq_folders():
                if watch_dir.is_uploaded(seq_folder.get_name()):
                    logging.info("{} already uploaded. Skipping".format(seq_folder.get_name()))
                    continue
                try:
                    uploader = DxUploaderFactory.create(self.settings, seq_folder)
                    uploader.upload(seq_folder)
                    watch_dir.complete(seq_folder.get_name())
                except UploaderException as e:
                    logging.error("{}. Skipping".format(e))
                    continue
            time.sleep(self.settings.get_interval())


class DxUploaderBase(object):
    __metaclass__ = ABCMeta

    def __init__(self, settings):
        self.settings = settings
        self.storage = StorageFactory.create(settings)

    def __call__(self, data_file):
        return self.upload_file(data_file)

    def upload(self, seq_folder):
        self._validate_target_dir(seq_folder)
        logging.info("Starting upload for {}".format(seq_folder.get_name()))
        sentinel = self._create_sentinel(seq_folder.get_name())
        if self.settings.get_process_count() > 1:
            with poolcontext(processes=self.settings.get_process_count()) as pool:
                results = pool.map(self, seq_folder.list_files())
        else:
            results = [self.upload_file(data_file) for data_file in seq_folder.list_files()]
        for data_file, file_id in results:
            sentinel.add_file(data_file, file_id)
        sentinel.close()
        logging.info("{} has been successfully uploaded".format(seq_folder.get_name()))

    def upload_file(self, data_file):
        remote_folder = os.path.join(self.settings.get_base_dir(),
                                     data_file.get_seq_folder_name(),
                                     data_file.get_relative_path()).replace("\\", "/")
        types = self._get_additional_types(data_file)
        types.append(Type.UPLOAD_DATA)
        properties = self._get_additional_properties(data_file, data_file.get_seq_folder_name())
        properties[Property.RUN_FOLDER] = data_file.get_seq_folder_name()
        file_id = self.storage.upload_file(data_file, remote_folder, types, properties)
        return data_file, file_id

    def _validate_target_dir(self, folder):
        self.storage.validate_target_dir(folder)

    @abstractmethod
    def _create_sentinel(self, seq_folder_name):
        raise NotImplementedError()

    @abstractmethod
    def _get_additional_types(self, data_file):
        raise NotImplementedError()

    @abstractmethod
    def _get_additional_properties(self, data_file, seq_folder_name):
        raise NotImplementedError()


class WesSignateraDxUploader(DxUploaderBase):
    SEQ_FOLDER_TYPE = WesSignateraSeqFolder

    def _create_sentinel(self, seq_folder_name):
        return WesSignateraSentinel(self.storage, self.settings.get_base_dir(), seq_folder_name)

    def _get_additional_types(self, data_file):
        types = []
        data_type = data_file.get_type()
        if data_type:
            types.append(data_type)
            if data_type == Type.CSV and data_file.get_name().startswith("WES-QCMetrics"):
                types.append(Type.WESQCREPORT)
        return types

    def _get_additional_properties(self, data_file, seq_folder_name):
        properties = {}
        if data_file.get_sample_id():
            properties[Property.SAMPLE_REFERENCE] = "{}/{}".format(seq_folder_name, data_file.get_sample_id())
        return properties


class PersonalisDxUploader(DxUploaderBase):
    SEQ_FOLDER_TYPE = PersonalisSeqFolder

    def _create_sentinel(self, seq_folder_name):
        return PersonalisSentinel(self.storage, self.settings.get_base_dir(), seq_folder_name)

    def _get_additional_types(self, data_file):
        types = []
        data_type = data_file.get_type()
        if data_type:
            types.append(data_type)
            if data_type == Type.CSV and data_file.get_name().startswith("QCMetrics"):
                types.append(Type.WESQCREPORT)
        return types

    def _get_additional_properties(self, data_file, seq_folder_name):
        properties = {}
        if data_file.get_sample_id():
            properties[Property.SAMPLE_REFERENCE] = "{}/{}".format(seq_folder_name, data_file.get_sample_id())
        return properties


class RawDxUploader(DxUploaderBase):
    SEQ_FOLDER_TYPE = RawSeqFolder

    def _create_sentinel(self, seq_folder_name):
        return RawSentinel(self.storage, self.settings.get_base_dir(), seq_folder_name)

    def _get_additional_types(self, data_file):
        return []

    def _get_additional_properties(self, data_file, seq_folder_name):
        return {}


class BgiDxUploader(DxUploaderBase):
    SEQ_FOLDER_TYPE = BgiSeqFolder

    def _create_sentinel(self, seq_folder_name):
        return RawSentinel(self.storage, self.settings.get_base_dir(), seq_folder_name)

    def _get_additional_types(self, data_file):
        return []

    def _get_additional_properties(self, data_file, seq_folder_name):
        return {}


class DxUploaderFactory(object):
    @staticmethod
    def create(settings, seq_folder):
        if isinstance(seq_folder, BgiDxUploader.SEQ_FOLDER_TYPE):
            return BgiDxUploader(settings)
        elif isinstance(seq_folder, RawDxUploader.SEQ_FOLDER_TYPE):
            return RawDxUploader(settings)
        elif isinstance(seq_folder, WesSignateraDxUploader.SEQ_FOLDER_TYPE):
            return WesSignateraDxUploader(settings)
        raise UploaderException(
                "Uploader for the folder {} was not found".format(seq_folder.get_name()))



#!/usr/bin/env python

import logging

from soyuz.configuration import Parameters
from soyuz.configuration import Settings
from soyuz.data.daemons import Daemon
from soyuz.data.folders import WesSignateraSeqFolder, BgiSeqFolder, RawSeqFolder, PersonalisSeqFolder, WatchDirectory
from soyuz.data.storages import StorageFactory
from soyuz.dx.uploaders import WesSignateraDxUploader, BgiDxUploader, RawDxUploader, PersonalisDxUploader, WatchUploader
from soyuz.utils import UploaderException


def main():
    try:
        params = Parameters()
        settings = Settings()

        settings.logging.initialize()

        if params.get_action() in [Parameters.UPLOAD_ACTION, Parameters.WATCH_ACTION]:

            settings.set_token(params.get_token() or settings.get_token())

            folder = None
            uploader = None

            if params.get_action() == Parameters.UPLOAD_ACTION:
                if params.get_upload_type() == Parameters.WES_SIGNATERA_UPLOAD:
                    uploader = WesSignateraDxUploader(settings)
                    folder = WesSignateraSeqFolder(params.get_folder())
                elif params.get_upload_type() == Parameters.BGI_UPLOAD:
                    uploader = BgiDxUploader(settings)
                    folder = BgiSeqFolder(params.get_folder())
                elif params.get_upload_type() == Parameters.RAW_UPLOAD:
                    uploader = RawDxUploader(settings)
                    folder = RawSeqFolder(params.get_folder())
                elif params.get_upload_type() == Parameters.PERSONALIS_UPLOAD:
                    uploader = PersonalisDxUploader(settings)
                    folder = PersonalisSeqFolder(params.get_folder())

                if not folder or not uploader:
                    raise UploaderException("Incorrect upload type")

                if not folder.is_valid:
                    raise UploaderException("Data folder is not in a valid state")

                uploader.upload(folder)

            elif params.get_action() == Parameters.WATCH_ACTION:

                settings.set_interval(params.get_interval() or settings.get_interval())

                uploader = WatchUploader(settings)

                if params.get_watch_type() == Parameters.START_WATCH:
                    if params.foreground():
                        uploader.watch(WatchDirectory(params.get_folder()))
                    else:
                        Daemon(uploader.watch, WatchDirectory(params.get_folder())).start()
                elif params.get_watch_type() == Parameters.STOP_WATCH:
                    Daemon().stop()
                elif params.get_watch_type() == Parameters.STATUS_WATCH:
                    Daemon().status()

        elif params.get_action() == Parameters.CONFIG_ACTION:

            if params.get_config_action() == Parameters.GET_CONFIG:
                if params.get_config_parameter_key() == "storage":
                    logging.info(settings.get_storage())
                elif params.get_config_parameter_key() == "token":
                    logging.info(settings.get_token())
                elif params.get_config_parameter_key() == "exodus_url":
                    logging.info(settings.get_exodus_url())
                elif params.get_config_parameter_key() == "basedir":
                    logging.info(settings.get_base_dir())
                elif params.get_config_parameter_key() == "interval":
                    logging.info(settings.get_interval())
                elif params.get_config_parameter_key() == "ua_path":
                    logging.info(settings.get_ua_path())
                elif params.get_config_parameter_key() == "ua_parameters":
                    logging.info(settings.get_ua_parameters())
                elif params.get_config_parameter_key() == "process_count":
                    logging.info(settings.get_process_count())
                elif params.get_config_parameter_key() == "chunk_size":
                    logging.info(settings.get_chunk_size())

            elif params.get_config_action() == Parameters.SET_CONFIG:
                if params.get_config_parameter_key() == "storage":
                    settings.set_storage(params.get_config_parameter_value())
                elif params.get_config_parameter_key() == "token":
                    settings.set_token(params.get_config_parameter_value())
                elif params.get_config_parameter_key() == "exodus_url":
                    settings.set_exodus_url(params.get_config_parameter_value())
                elif params.get_config_parameter_key() == "exodus_username":
                    settings.set_exodus_username(params.get_config_parameter_value())
                elif params.get_config_parameter_key() == "exodus_password":
                    settings.set_exodus_password(params.get_config_parameter_value())
                elif params.get_config_parameter_key() == "basedir":
                    settings.set_basedir(params.get_config_parameter_value())
                elif params.get_config_parameter_key() == "interval":
                    settings.set_interval(params.get_config_parameter_value())
                elif params.get_config_parameter_key() == "ua_path":
                    settings.set_ua_path(params.get_config_parameter_value())
                elif params.get_config_parameter_key() == "ua_parameters":
                    settings.set_ua_parameters(params.get_config_parameter_value())
                elif params.get_config_parameter_key() == "process_count":
                    settings.set_process_count(params.get_config_parameter_value())
                elif params.get_config_parameter_key() == "chunk_size":
                    settings.set_chunk_size(params.get_config_parameter_value())
                else:
                    raise UploaderException("There is no parameter '{}'".format(params.get_config_parameter_key()))
                settings.dump()
                logging.info("Set {} to {}".format(params.get_config_parameter_key(), params.get_config_parameter_value()))

    except UploaderException as e:
        logging.error(e)
        quit(1)


if __name__ == "__main__":
    main()
